from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions
    from mipac.types import FolderPayload, IDriveFile, IFileProperties

__all__ = ['FileProperties', 'File', 'Folder']


class FileProperties:
    def __init__(self, properties: IFileProperties) -> None:
        self.__properties: IFileProperties = properties

    @property
    def width(self) -> Optional[int]:
        return self.__properties['width']

    @property
    def height(self) -> int:
        return self.__properties['height']

    @property
    def avg_color(self) -> Optional[str]:
        return self.__properties['avg_color']


class Folder:
    def __init__(self, folder: FolderPayload, client: ClientActions):
        self.__folder: FolderPayload = folder
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        """フォルダのID"""
        return self.__folder['id']

    @property
    def created_at(self) -> str:  # TODO: 型
        """フォルダの作成日時"""
        return self.__folder['created_at']

    @property
    def name(self) -> str:
        """フォルダ名"""
        return self.__folder['name']

    @property
    def folders_count(self) -> int:
        """フォルダ内のフォルダ数"""
        return self.__folder['folders_count']

    @property
    def files_count(self) -> int:
        """フォルダ内のファイル数"""
        return self.__folder['files_count']

    @property
    def parent_id(self) -> str:
        return self.__folder['parent_id']

    @property
    def parent(self) -> dict[str, Any]:
        return self.__folder['parent']


class File:
    def __init__(self, file: IDriveFile, *, client: ClientActions):
        self.__file: IDriveFile = file
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__file['id']

    @property
    def created_at(self):
        return self.__file['created_at']

    @property
    def is_sensitive(self) -> bool:
        return self.__file['is_sensitive']

    @property
    def name(self) -> str:
        return self.__file['name']

    @property
    def thumbnail_url(self) -> str:
        return self.__file['thumbnail_url']

    @property
    def url(self) -> str:
        return self.__file['url']

    @property
    def type(self) -> str:
        return self.__file['type']

    @property
    def size(self) -> int:
        return self.__file['size']

    @property
    def md5(self) -> str:
        return self.__file['md5']

    @property
    def blurhash(self) -> str:
        return self.__file['blurhash']

    @property
    def properties(self) -> FileProperties:
        return FileProperties(self.__file['properties'])



from .drive import File, FileProperties, Folder
from .note import Follow, Header, Note, NoteReaction, Poll
from .notification import Reaction
from .user import FollowRequest, UserDetailed

__all__ = (
    'FileProperties',
    'File',
    'Folder',
    'UserDetailed',
    'FollowRequest',
    'Note',
    'Poll',
    'Reaction',
    'Follow',
    'Header',
    'NoteReaction',
)



from typing import Optional, TypedDict

__all__ = ('EmojiPayload', 'ICustomEmojiLite', 'ICustomEmoji')


class ICustomEmojiLite(TypedDict):
    name: str
    url: str


class ICustomEmoji(ICustomEmojiLite):
    id: str
    category: str
    aliases: list[str]


class EmojiPayload(TypedDict):
    id: Optional[str]
    aliases: Optional[list[str]]
    name: Optional[str]
    category: Optional[str]
    host: Optional[str]
    url: Optional[str]



"""
WebSocketで来る通知の型を定義しています

NfはNotificationの略です
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from mipac.types.note import INote
    from mipac.types.user import IUserLite


class Notification(TypedDict):
    id: str
    created_at: str
    is_read: bool


class IReactionNf(Notification):
    type: str
    reaction: str
    user: IUserLite
    user_id: str
    note: INote



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal, TypedDict

if TYPE_CHECKING:
    from mipac.types.channel import IChannel
    from mipac.types.emoji import ICustomEmojiLite
    from mipac.types.instance import IInstanceLite
    from mipac.types.note import INote
    from mipac.types.page import IPage
__all__ = (
    'IChannel',
    'IUserLite',
    'IUserDetailed',
    'IUserDetailedField',
)


class IUserLite(TypedDict):
    id: str
    username: str
    host: str | None
    name: str
    online_status: Literal['online', 'active', 'offline', 'unknown']
    avatar_url: str
    avatar_blurhash: str
    emojis: list[ICustomEmojiLite]
    instance: IInstanceLite


class IUserDetailedField(TypedDict):
    name: str
    value: str


class IUserDetailedRequired(IUserLite):
    fields: list[IUserDetailedField]
    followers_count: int
    following_count: int
    has_pending_follow_request_from_you: bool
    has_pending_follow_request_to_you: bool
    is_admin: bool
    is_blocked: bool
    is_blocking: bool
    is_bot: bool
    is_cat: bool
    is_followed: bool
    is_following: bool
    is_locked: bool
    is_moderator: bool
    is_muted: bool
    is_silenced: bool
    is_suspended: bool
    public_reactions: bool
    security_keys: bool
    two_factor_enabled: bool
    notes_count: int
    pinned_note_ids: list[str]
    pinned_notes: List[INote]


class IUserDetailed(IUserDetailedRequired, total=False):
    banner_blurhash: str
    banner_color: str
    banner_url: str
    birthday: str
    created_at: str
    description: str
    ff_visibility: Literal['public', 'followers', 'private']
    lang: str
    last_fetched_at: str
    location: str
    pinned_page: IPage
    pinned_page_id: str
    updated_at: str
    uri: str
    url: str


class IFollowRequest(TypedDict):
    id: str
    follower: IUserLite
    followee: IUserLite



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.types.poll import IPoll, IPollChoice

if TYPE_CHECKING:
    from mipac.manager import ClientActions


class PollChoice:
    def __init__(self, choice: IPollChoice, *, client: ClientActions):
        self.__choice: IPollChoice = choice
        self.__client: ClientActions = client

    @property
    def is_voted(self) -> bool:
        return self.__choice['is_voted']

    @property
    def text(self) -> str:
        return self.__choice['text']

    @property
    def votes(self) -> int:
        return self.__choice['votes']


class Poll:
    def __init__(self, poll: IPoll, *, client: ClientActions):
        self.__poll: IPoll = poll
        self.__client: ClientActions = client

    @property
    def multiple(self) -> bool:
        return self.__poll['multiple']

    @property
    def expires_at(self) -> int:
        return self.__poll['expires_at']

    @property
    def choices(self) -> list[PollChoice]:
        return [
            PollChoice(i, client=self.__client) for i in self.__poll['choices']
        ]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac.abc.manager import AbstractManager
from mipac.actions.drive import DriveActions, FileActions, FolderActions
from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ['FolderManager', 'FileManager', 'DriveManager']


class FileManager(AbstractManager):
    def __init__(
        self,
        file_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.__file_id = file_id

    @property
    def action(self) -> FileActions:
        """
        ファイルの操作を行うインスタンスを返します

        Return
        ------
        FileActions
            ファイルに対するアクション
        """
        return FileActions(
            file_id=self.__file_id,
            client=self.__client,
            session=self.__session,
        )


class FolderManager(AbstractManager):
    def __init__(
        self,
        folder_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__folder_id = folder_id
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.file: FileManager = FileManager(session=session, client=client)

    @property
    def action(self) -> FolderActions:
        """
        フォルダーの操作を行うインスタンスを返します

        Returns
        -------
        FolderActions
            フォルダーに対するアクション
        """
        return FolderActions(
            folder_id=self.__folder_id,
            session=self.__session,
            client=self.__client,
        )

    def _get_file_instance(self, file_id: str) -> FileManager:
        return FileManager(
            file_id=file_id, session=self.__session, client=self.__client
        )


class DriveManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.folder: FolderManager = FolderManager(
            session=session, client=client
        )
        self.file: FileManager = FileManager(session=session, client=client)

    @property
    def action(self) -> DriveActions:
        """
        ドライブの操作を行うインスタンスを返します

        Returns
        -------
        DriveActions
            ドライブに対するアクション
        """
        return DriveActions(client=self.__client, session=self.__session)

    def _get_folder_instance(self, folder_id: str) -> FolderManager:
        return FolderManager(
            session=self.__session, client=self.__client, folder_id=folder_id
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional


from mipac.exception import ParameterError
from mipac.http import HTTPClient, Route
from mipac.manager.favorite import FavoriteManager
from mipac.manager.file import MiFile
from mipac.manager.reaction import ReactionManager
from mipac.models.note import Note, NoteReaction
from mipac.models.poll import Poll
from mipac.types.note import INote

__all__ = ['NoteActions']

from mipac.util import (
    check_multi_arg,
    remove_dict_empty,
    cache,
)

if TYPE_CHECKING:
    from mipac.client import ClientActions


class NoteActions:
    def __init__(
        self,
        note_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__note_id: Optional[str] = note_id
        self.favorite = FavoriteManager(
            note_id=note_id, session=session, client=client
        )
        self.reaction = ReactionManager(
            note_id=note_id, session=session, client=client
        )
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def add_clips(
        self, clip_id: str, note_id: Optional[str] = None
    ) -> bool:
        """
        クリップに追加します

        Parameters
        ----------
        note_id : Optional[str], default=None
                追加するノートのID
        clip_id : str
            クリップのID
        note_id : Optional[str], default=None
            追加したいノートのID

        Returns
        -------
        bool
            成功したか否か
        """

        note_id = note_id or self.__note_id

        data = {'noteId': note_id, 'clipId': clip_id}
        return bool(
            await self.__session.request(
                Route('POST', '/api/clips/add-note'), json=data, auth=True
            )
        )

    async def send(
        self,
        content: Optional[str] = None,
        visibility: str = 'public',
        visible_user_ids: Optional[list[str]] = None,
        cw: Optional[str] = None,
        local_only: bool = False,
        extract_mentions: bool = True,
        extract_hashtags: bool = True,
        extract_emojis: bool = True,
        reply_id: Optional[str] = None,
        renote_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        files: Optional[list[MiFile]] = None,
        poll: Optional[Poll] = None,
    ) -> Note:
        """
        ノートを投稿します。

        Parameters
        ----------
        content : Optional[str], default=None
            投稿する内容
        visibility : str, optional
            公開範囲, by default "public"
            Enum: "public" "home" "followers" "specified"
        visible_user_ids : Optional[list[str]], optional
            公開するユーザー, by default None
        cw : Optional[str], optional
            閲覧注意の文字, by default None
        local_only : bool, optional
            ローカルにのみ表示するか, by default False
        extract_mentions : bool, optional
            メンションを展開するか, by default False
        extract_hashtags : bool, optional
            ハッシュタグを展開するか, by default False
        extract_emojis : bool, optional
            絵文字を展開するか, by default False
        reply_id : Optional[str], optional
            リプライ先のid, by default None
        renote_id : Optional[str], optional
            リノート先のid, by default None
        channel_id : Optional[str], optional
            チャンネルid, by default None
        files : list[MiFile], optional
            添付するファイルのリスト, by default None
        poll : Optional[Poll], optional
            アンケート, by default None

        Returns
        -------
        Note
            投稿したノート

        Raises
        ------
        ContentRequired
            [description]
        """
        file_ids = None
        if files:
            file_ids = [file.file_id for file in files]

        field = {
            'visibility': visibility,
            'visibleUserIds': visible_user_ids,
            'text': content,
            'cw': cw,
            'localOnly': local_only,
            'noExtractMentions': not extract_mentions,
            'noExtractHashtags': not extract_hashtags,
            'noExtractEmojis': not extract_emojis,
            'replyId': reply_id,
            'renoteId': renote_id,
            'channelId': channel_id,
            'fileIds': file_ids,
        }
        if not check_multi_arg(content, files, renote_id, poll):
            raise ParameterError(
                'ノートの送信にはcontent, file_ids, renote_id またはpollのいずれか1つが無くてはいけません'
            )

        if poll and type(Poll):
            poll_data = remove_dict_empty(
                {
                    'choices': poll.choices,
                    'multiple': poll.multiple,
                    'expiresAt': poll.expires_at,
                    'expiredAfter': poll.expired_after,
                }
            )
            field['poll'] = poll_data

        # if files:  # TODO: get_file_idsを直さないと使えない
        # field['fileIds'] = await get_file_ids(files=files)
        field = remove_dict_empty(field)
        res: INote = await self.__session.request(
            Route('POST', '/api/notes/create'),
            json=field,
            auth=True,
            lower=True,
        )
        return Note(res, client=self.__client)

    async def delete(self, note_id: Optional[str] = None) -> bool:
        """
        ノートを削除します

        Parameters
        ----------
        note_id : Optional[str], default=None
            削除したいノートのID

        Returns
        -------
        bool
            削除に成功したか否か
        """

        note_id = note_id or self.__note_id

        data = {'noteId': note_id}
        res = await self.__session.request(
            Route('POST', '/api/notes/delete'), json=data, auth=True
        )
        return bool(res)

    async def create_renote(self, note_id: Optional[str] = None) -> Note:
        """
        リノートを作成します

        Parameters
        ----------
        note_id : Optional[str], default=None
            ノートのID

        Returns
        -------
        Note
            作成したリノート
        """
        note_id = note_id or self.__note_id
        return await self.send(renote_id=note_id)

    async def create_quote(
        self,
        content: Optional[str] = None,
        visibility: str = 'public',
        visible_user_ids: Optional[list[str]] = None,
        cw: Optional[str] = None,
        local_only: bool = False,
        extract_mentions: bool = True,
        extract_hashtags: bool = True,
        extract_emojis: bool = True,
        file_ids: Optional[list[str]] = None,
        poll: Optional[Poll] = None,
        note_id: Optional[str] = None,
    ) -> Note:
        """
        Create a note quote.

        Parameters
        ----------
        content: Optional[str], default=None
            text
        visibility: str, default='public'
            Disclosure range
        visible_user_ids: Optional[list[str]], default=None
            List of users to be published
        cw: Optional[str], default=None
            Text to be displayed when warning is given
        local_only: bool, default=False
            Whether to show only locally or not
        extract_mentions: bool, default=True
            Whether to expand the mention
        extract_hashtags: bool, default=True
            Whether to expand the hashtag
        extract_emojis: bool, default=True
            Whether to expand the emojis
        file_ids: Optional[list[str]], default=None
            The ID list of files to be attached
        poll: Optional[Poll], default=None
            Questionnaire to be created
        note_id: Optional[str], default=None
            Note IDs to target for renote and citations
        """

        note_id = note_id or self.__note_id

        return await self.send(
            content=content,
            visibility=visibility,
            visible_user_ids=visible_user_ids,
            cw=cw,
            local_only=local_only,
            extract_mentions=extract_mentions,
            extract_hashtags=extract_hashtags,
            extract_emojis=extract_emojis,
            renote_id=note_id,
            poll=poll,
        )  # TODO: filesを受け取るように

    @cache(group='get_note')
    async def get(self, note_id: Optional[str] = None, **kwargs) -> Note:
        """
        ノートを取得します

        Parameters
        ----------
        note_id : Optional[str], default=None
            ノートのID

        Returns
        -------
        Note
            取得したノートID
        """
        note_id = note_id or self.__note_id
        res = await self.__session.request(
            Route('POST', '/api/notes/show'),
            json={'noteId': note_id},
            auth=True,
            lower=True,
        )
        return Note(res, client=self.__client)

    @cache(group='get_note', override=True)
    async def fetch(self, note_id: Optional[str] = None) -> Note:
        note_id = note_id or self.__note_id
        res = await self.__session.request(
            Route('POST', '/api/notes/show'),
            json={'noteId': note_id},
            auth=True,
            lower=True,
        )
        return Note(res, client=self.__client)

    async def get_replies(
        self,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        limit: int = 10,
        note_id: Optional[str] = None,
    ) -> list[Note]:
        """
        ノートに対する返信を取得します

        Parameters
        ---------
        since_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより新しい投稿を取得します
        until_id : Optional[str], default=None
            指定すると、その投稿を投稿を起点としてより古い投稿を取得します
        limit : int, default=10
            取得する上限
        note_id: Optional[str], default=None
            返信を取得したいノートのID

        Returns
        -------
        list[Note]
            返信のリスト
        """
        note_id = note_id or self.__note_id
        res = await self.__session.request(
            Route('POST', '/api/notes/replies'),
            json={
                'noteId': note_id,
                'sinceId': since_id,
                'untilId': until_id,
                'limit': limit,
            },
            auth=True,
            lower=True,
        )
        return [Note(i, client=self.__client) for i in res]

    async def get_reaction(
        self, reaction: str, note_id: Optional[str] = None
    ) -> list[NoteReaction]:
        note_id = note_id or self.__note_id
        return await ReactionManager(
            note_id=note_id, client=self.__client, session=self.__session
        ).get_reaction(reaction)
        # TODO: self.__clientを使ったインスタンス生成に変えないと循環インポートの原因になりかねない



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mipac.models.emoji import CustomEmoji
from mipac.types.ads import IAds
from mipac.types.instance import IInstanceLite, IInstanceMetaLite

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions


class LiteInstanceMeta:
    def __init__(
        self, instance: IInstanceMetaLite, *, client: ClientActions
    ) -> None:
        self.__instance: IInstanceMetaLite = instance
        self.__client: ClientActions = client

    @property
    def version(self) -> str:
        return self.__instance['version']

    @property
    def uri(self) -> str:
        return self.__instance['uri']

    @property
    def disable_registration(self) -> bool:
        return self.__instance['disable_registration']

    @property
    def disable_local_timeline(self) -> bool:
        return self.__instance['disable_local_timeline']

    @property
    def disable_global_timeline(self) -> bool:
        return self.__instance['disable_global_timeline']

    @property
    def drive_capacity_per_local_user_mb(self) -> int:
        return self.__instance['drive_capacity_per_local_user_mb']

    @property
    def drive_capacity_per_remote_user_mb(self) -> int:
        return self.__instance['drive_capacity_per_remote_user_mb']

    @property
    def enable_hcaptch(self) -> bool:
        return self.__instance['enable_hcaptcha']

    @property
    def max_note_text_length(self) -> int:
        return self.__instance['max_note_text_length']

    @property
    def enable_email(self) -> bool:
        return self.__instance['enable_email']

    @property
    def enable_twitter_integration(self) -> bool:
        return self.__instance['enable_twitter_integration']

    @property
    def enable_github_integration(self) -> bool:
        return self.__instance['enable_github_integration']

    @property
    def enable_discord_integration(self) -> bool:
        return self.__instance['enable_discord_integration']

    @property
    def enable_service_worker(self) -> bool:
        return self.__instance['enable_service_worker']

    @property
    def emojis(self) -> list[CustomEmoji]:
        return [
            CustomEmoji(i, client=self.__client)
            for i in self.__instance['emojis']
        ]

    @property
    def mascot_image_url(self) -> str:
        return self.__instance['mascot_image_url']

    @property
    def banner_url(self) -> str:
        return self.__instance['banner_url']

    @property
    def icon_url(self) -> str:
        return self.__instance['icon_url']

    @property
    def maintainer_name(self) -> str | None:
        return self.__instance.get('maintainer_name')

    @property
    def maintainer_email(self) -> str | None:
        return self.__instance.get('maintainer_email')

    @property
    def name(self) -> str | None:
        return self.__instance.get('name')

    @property
    def description(self) -> str | None:
        return self.__instance.get('description')

    @property
    def langs(self) -> list[str]:
        return self.__instance.get('langs', [])

    @property
    def tos_url(self) -> str | None:
        return self.__instance.get('tos_url')

    @property
    def tos_text_url(self) -> str | None:
        return self.__instance.get('tos_text_url')

    @property
    def announcements(self) -> dict[str, Any] | None:  # TODO: 型確認
        return self.__instance.get('announcements')

    @property
    def hcaptcha_site_key(self) -> str | None:
        return self.__instance.get('hcaptcha_site_key')

    @property
    def enable_recaptcha(self) -> bool:
        return self.__instance.get('enable_recaptcha', False)

    @property
    def recaptcha_siteKey(self) -> str | None:
        return self.__instance.get('recaptcha_siteKey')

    @property
    def sw_publickey(self) -> str | None:
        return self.__instance.get('sw_publickey')

    @property
    def ads(self) -> list[IAds] | None:  # TODO: モデルに
        return self.__instance.get('ads')


class LiteInstance:
    def __init__(self, instance: IInstanceLite) -> None:
        self.__instance: IInstanceLite = instance

    @property
    def name(self) -> str:
        return self.__instance['name']

    def software_name(self) -> str:
        return self.__instance['software_name']

    def software_version(self) -> str:
        return self.__instance['software_version']

    def icon_url(self) -> str:
        return self.__instance['icon_url']

    def favicon_url(self) -> str:
        return self.__instance['favicon_url']

    def theme_color(self) -> str:
        return self.__instance['theme_color']



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac.http import HTTPClient, Route
from mipac.manager.ad import AdminAdvertisingManager
from mipac.manager.admin.user import AdminUserManager
from mipac.manager.emoji import AdminEmojiManager
from mipac.manager.moderator import AdminModeratorManager

if TYPE_CHECKING:
    from mipac.client import ClientActions


class AdminActions:
    def __init__(self, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.emoji = AdminEmojiManager(session=session, client=client)
        self.user = AdminUserManager(session=session, client=client)
        self.ad = AdminAdvertisingManager(session=session, client=client)
        self.moderator = AdminModeratorManager(session=session, client=client)

    def get_emoji_instance(
        self, emoji_id: Optional[str] = None
    ) -> AdminEmojiManager:
        return AdminEmojiManager(
            emoji_id, session=self.__session, client=self.__client
        )

    async def get_invite(self) -> bool:
        return bool(
            await self.__session.request(Route('POST', '/api/admin/invite'))
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.models.drive import File
from mipac.models.lite.user import LiteUser

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions
    from mipac.types.chat import IChatGroup, IChatMessage
    from mipac.actions.chat import BaseChatAction

__all__ = ['ChatGroup', 'ChatMessage']


class ChatGroup:
    def __init__(self, group: IChatGroup, *, client: ClientActions):
        self.__group: IChatGroup = group
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        """グループのID"""
        return self.__group['id']

    @property
    def created_at(self) -> str:
        """グループの作成日時"""
        return self.__group['created_at']

    @property
    def name(self) -> str:
        """グループ名"""
        return self.__group['name']

    @property
    def owner_id(self) -> str:
        """グループのオーナーのID"""
        return self.__group['owner_id']

    @property
    def user_ids(self) -> list[str]:
        return self.__group['user_ids']


class ChatMessage:
    """
    チャットオブジェクト
    """

    def __init__(self, chat: IChatMessage, *, client: ClientActions):
        self.__chat: IChatMessage = chat
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__chat['id']

    @property
    def created_at(self) -> str:
        return self.__chat['created_at']

    @property
    def file(self) -> File:
        return File(self.__chat['file'], client=self.__client)

    @property
    def text(self) -> str | None:
        return self.__chat['text']

    @property
    def user_id(self) -> str:
        return self.__chat['user_id']

    @property
    def user(self) -> LiteUser:
        return LiteUser(self.__chat['user'])

    @property
    def recipient_id(self) -> str:
        return self.__chat['recipient_id']

    @property
    def recipient(self) -> str:
        return self.__chat['recipient']

    @property
    def group_id(self) -> str:
        return self.__chat['group_id']

    @property
    def file_id(self) -> str:
        return self.__chat['file_id']

    @property
    def is_read(self) -> bool:
        return bool(self.__chat['is_read'])

    @property
    def reads(self) -> list[str]:
        return self.__chat['reads']

    @property
    def group(self) -> ChatGroup | None:
        return (
            ChatGroup(self.__chat['group'], client=self.__client)
            if self.__chat['group']
            else None
        )

    @property
    def action(self) -> BaseChatAction:
        return self.__client.chat.custom_base_chat_action(
            user_id=self.user.id, message_id=self.id
        )



"""
MiPAを使用する上でちょっとした際に便利なツール一覧
"""
from __future__ import annotations

import asyncio
import json
import re
import uuid
from _operator import itemgetter
from datetime import datetime, timedelta
from typing import Any, Optional
from urllib.parse import urlencode
import functools

import aiohttp

try:
    import orjson  # type: ignore
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True

__all__ = (
    'deprecated_func',
    'MiTime',
    'get_cache_key',
    'key_builder',
    'check_multi_arg',
    'remove_list_empty',
    'remove_dict_empty',
    'upper_to_lower',
    'str_lower',
    'bool_to_string',
    '_from_json',
)


if HAS_ORJSON:
    _from_json = orjson.loads  # type: ignore
else:
    _from_json = json.loads

DEFAULT_CACHE: dict[str, list[dict[str, Any]]] = {}


def deprecated_func(func):
    print(f'deprecated function:{func.__name__}')


class MiTime:
    def __init__(self, start: timedelta, end: datetime):
        self.start = start
        self.end = end


class AuthClient:
    """
    Tokenの取得を手助けするクラス
    """

    def __init__(
        self,
        instance_uri: str,
        name: str,
        description: str,
        permissions: Optional[list[str]] = None,
        *,
        icon: Optional[str] = None,
        use_miauth: bool = False,
    ):
        """
        Parameters
        ----------
        instance_uri : str
            アプリケーションを作成したいインスタンスのURL
        name : str
            アプリケーションの名前
        description : str
            アプリケーションの説明
        permissions : Optional[list[str]], default=None
            アプリケーションが要求する権限
        icon: Optional[str], default=None
            アプリケーションのアイコン画像URL
        use_miauth: bool, default=False
            MiAuthを使用するか
        """
        if permissions is None:
            permissions = ['read:account']
        self.__client_session = aiohttp.ClientSession()
        self.__instance_uri: str = instance_uri
        self.__name: str = name
        self.__description: str = description
        self.__permissions: list[str] = permissions
        self.__icon: Optional[str] = icon
        self.__use_miauth: bool = use_miauth
        self.__session_token: uuid.UUID
        self.__secret: str

    async def get_auth_url(self) -> str:
        """
        認証に使用するURLを取得します

        Returns
        -------
        str
            認証に使用するURL
        """
        field = remove_dict_empty(
            {
                'name': self.__name,
                'description': self.__description,
                'icon': self.__icon,
            }
        )
        if self.__use_miauth:
            field['permissions'] = self.__permissions
            query = urlencode(field)
            self.__session_token = uuid.uuid4()
            return (
                f'{self.__instance_uri}/miauth/{self.__session_token}?{query}'
            )
        else:
            field['permission'] = self.__permissions
            async with self.__client_session.post(
                f'{self.__instance_uri}/api/app/create', json=field
            ) as res:
                data = await res.json()
                self.__secret = data['secret']
            async with self.__client_session.post(
                f'{self.__instance_uri}/api/auth/session/generate',
                json={'appSecret': self.__secret},
            ) as res:
                data = await res.json()
                self.__session_token = data['token']
                return data['url']

    async def wait_miauth(self) -> str:
        url = f'{self.__instance_uri}/api/miauth/{self.__session_token}/check'
        while True:
            async with self.__client_session.post(url) as res:
                data = await res.json()
                if data.get('ok') is True:
                    return data
            await asyncio.sleep(1)

    async def wait_oldauth(self) -> None:
        while True:
            async with self.__client_session.post(
                f'{self.__instance_uri}/api/auth/session/userkey',
                json={
                    'appSecret': self.__secret,
                    'token': self.__session_token,
                },
            ) as res:
                data = await res.json()
                if data.get('error', {}).get('code') != 'PENDING_SESSION':
                    break
            await asyncio.sleep(1)

    async def check_auth(self) -> str:
        """
        認証が完了するまで待機し、完了した場合はTokenを返します

        Returns
        -------
        str
            Token
        """
        if self.__use_miauth:
            data = await self.wait_miauth()
        else:
            data = await self.wait_oldauth()
        await self.__client_session.close()
        return data['token'] if self.__use_miauth else data['accessToken']


def dynamic_args(decorator):
    def wrapper(*args, **kwargs):
        if len(args) != 0 and callable(args[0]):
            func = args[0]
            return functools.wraps(func)(decorator(func))
        else:

            def _wrapper(func):
                return functools.wraps(func)(decorator(func, *args, **kwargs))

            return _wrapper

    return wrapper


def set_cache(group: str, key: str, value: Any):
    if len(DEFAULT_CACHE.get(group, [])) > 50:
        del DEFAULT_CACHE[group][-1]

    if DEFAULT_CACHE.get(group) is None:
        DEFAULT_CACHE[group] = []
    DEFAULT_CACHE[group].append({key: value})


def get_cache(group: str, key: str):
    if DEFAULT_CACHE.get(group):
        return list(map(itemgetter(key), DEFAULT_CACHE[group]))[0]
    return None


@dynamic_args
def cache(func, group: str = 'default', override: bool = False):
    async def decorator(self, *args, **kwargs):
        ordered_kwargs = sorted(kwargs.items())
        key = '.{0}' + str(args) + str(ordered_kwargs)
        hit_item = get_cache(group, key)
        print(hit_item)
        if hit_item and override is False:
            return hit_item
        res = await func(self, *args, **kwargs)
        set_cache(group, key, res)
        return res

    return decorator


def get_cache_key(func):
    async def decorator(self, *args, **kwargs):
        ordered_kwargs = sorted(kwargs.items())
        key = (
            (func.__module__ or '')
            + '.{0}'
            + f'{self}'
            + str(args)
            + str(ordered_kwargs)
        )
        return await func(self, *args, **kwargs, cache_key=key)

    return decorator


def key_builder(func, cls, *args, **kwargs):
    ordered_kwargs = sorted(kwargs.items())
    key = (
        (func.__module__ or '')
        + f'.{func.__name__}'
        + f'{cls}'
        + str(args)
        + str(ordered_kwargs)
    )
    return key


def check_multi_arg(*args: Any) -> bool:
    """
    複数の値を受け取り値が存在するかをboolで返します

    Parameters
    ----------
    args : list
        確認したい変数のリスト

    Returns
    -------
    bool
        存在する場合はTrue, 存在しない場合はFalse
    """
    return bool([i for i in args if i])


def remove_list_empty(data: list[Any]) -> list[Any]:
    """
    Parameters
    ----------
    data: dict
        空のkeyを削除したいdict

    Returns
    -------
    dict[str, Any]
        空のkeyがなくなったdict
    """
    return [k for k in data if k]


def remove_dict_empty(data: dict[str, Any]) -> dict[str, Any]:
    """
    Parameters
    ----------
    data: dict
        空のkeyを削除したいdict

    Returns
    -------
    _data: dict
        空のkeyがなくなったdict
    """
    _data = {}
    _data = {k: v for k, v in data.items() if v is not None}
    _data = {k: v for k, v in data.items() if v}
    return _data


def upper_to_lower(
    data: dict[str, Any],
    field: Optional[dict[str, Any]] = None,
    nest: bool = True,
    replace_list: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    """
    Parameters
    ----------
    data: dict
        小文字にしたいkeyがあるdict
    field: dict, default=None
        謎
    nest: bool, default=True
        ネストされたdictのkeyも小文字にするか否か
    replace_list: dict, default=None
        dictのkey名を特定の物に置き換える

    Returns
    -------
    field : dict
        小文字になった, key名が変更されたdict
    """
    if data is None:
        return {}
    if replace_list is None:
        replace_list = {}

    if field is None:
        field = {}
    for attr in data:
        pattern = re.compile('[A-Z]')
        large = [i.group().lower() for i in pattern.finditer(attr)]
        result = [None] * (len(large + pattern.split(attr)))
        result[::2] = pattern.split(attr)
        result[1::2] = ['_' + i.lower() for i in large]
        default_key = ''.join(result)
        if replace_list.get(attr):
            default_key = default_key.replace(attr, replace_list.get(attr))
        field[default_key] = data[attr]
        if isinstance(field[default_key], dict) and nest:
            field[default_key] = upper_to_lower(field[default_key])
    return field


def str_lower(text: str):
    pattern = re.compile('[A-Z]')
    large = [i.group().lower() for i in pattern.finditer(text)]
    result = [None] * (len(large + pattern.split(text)))
    result[::2] = pattern.split(text)
    result[1::2] = ['_' + i.lower() for i in large]
    return ''.join(result)


def bool_to_string(boolean: bool) -> str:
    """
    boolを小文字にして文字列として返します

    Parameters
    ----------
    boolean : bool
        変更したいbool値
    Returns
    -------
    true or false: str
        小文字になったbool文字列
    """
    return 'true' if boolean else 'false'



from __future__ import annotations

from mipac.types.chart import (
    ActiveUsersChartPayload,
    DriveChartPayload,
    DriveLocalChartPayload,
    DriveRemoteChartPayload,
)

__all__ = (
    'RawActiveUsersChart',
    'RawDriveRemoteChart',
    'RawDriveLocalChart',
    'RawDriveChart',
)


class RawActiveUsersChart:
    __slots__ = (
        'read_write',
        'read',
        'write',
        'registered_within_week',
        'registered_within_month',
        'registered_within_year',
        'registered_outside_week',
        'registered_outside_month',
        'registered_outside_year',
    )

    def __init__(self, data: ActiveUsersChartPayload):
        self.read_write: list[int] = data['read_write']
        self.read: list[int] = data['read']
        self.write: list[int] = data['write']
        self.registered_within_week: list[int] = data['registered_within_week']
        self.registered_within_month: list[int] = data[
            'registered_within_month'
        ]
        self.registered_within_year: list[int] = data['registered_within_year']
        self.registered_outside_week: list[int] = data[
            'registered_outside_week'
        ]
        self.registered_outside_month: list[int] = data[
            'registered_outside_month'
        ]
        self.registered_outside_year: list[int] = data[
            'registered_outside_year'
        ]


class RawDriveLocalChart:
    __slots__ = (
        'total_count',
        'total_size',
        'inc_count',
        'inc_size',
        'dec_count',
        'dec_size',
    )

    def __init__(self, data: DriveLocalChartPayload):
        self.total_count: list[int] = data['total_count']
        self.total_size: list[int] = data['total_size']
        self.inc_count: list[int] = data['inc_count']
        self.inc_size: list[int] = data['inc_size']
        self.dec_count: list[int] = data['dec_count']
        self.dec_size: list[int] = data['dec_size']


class RawDriveRemoteChart:
    __slots__ = (
        'total_count',
        'total_size',
        'inc_count',
        'inc_size',
        'dec_count',
        'dec_size',
    )

    def __init__(self, data: DriveRemoteChartPayload):
        self.total_count: list[int] = data['total_count']
        self.total_size: list[int] = data['total_size']
        self.inc_count: list[int] = data['inc_count']
        self.inc_size: list[int] = data['inc_size']
        self.dec_count: list[int] = data['dec_count']
        self.dec_size: list[int] = data['dec_size']


class RawDriveChart:
    __slots__ = ('local', 'remote')

    def __init__(self, data: DriveChartPayload):
        self.local: RawDriveLocalChart = RawDriveLocalChart(data['local'])
        self.remote: RawDriveRemoteChart = RawDriveRemoteChart(data['remote'])



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from mipac.manager.client import ClientActions
from mipac.manager.reaction import ReactionManager
from mipac.models.lite.user import LiteUser
from mipac.models.note import Note

if TYPE_CHECKING:
    from mipac.types.notification import IReactionNf


class Reaction:
    def __init__(
        self, reaction: IReactionNf, *, client: ClientActions
    ) -> None:
        self.__reaction: IReactionNf = reaction
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__reaction['id']

    @property
    def created_at(self) -> datetime:
        return datetime.strptime(
            self.__reaction['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )

    @property
    def type(self) -> str:
        return self.__reaction['type']

    @property
    def is_read(self) -> bool:
        return self.__reaction['is_read']

    @property
    def user(self) -> LiteUser:
        return LiteUser(self.__reaction['user'])

    @property
    def note(self) -> Note:
        return Note(self.__reaction['note'], client=self.__client)

    @property
    def reaction(self) -> str:
        return self.__reaction['reaction']

    @property
    def action(self) -> ReactionManager:
        return self.__client.reaction



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac.abc.manager import AbstractManager
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.client import ClientActions


class PagesManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def get_pages(
        self,
        limit: int = 100,
        since_id: Optional[int] = None,
        until_id: Optional[int] = None,
    ):
        data = {'limit': limit, 'since_id': since_id, 'until_id': until_id}
        res = await self.__session.request(
            Route('POST', '/api/i/pages'), json=data, auth=True
        )
        return res



from mipac.types.chart import (
    ActiveUsersChartPayload,
    DriveChartPayload,
    DriveLocalChartPayload,
    DriveRemoteChartPayload,
)
from mipac.types.drive import FolderPayload, IDriveFile, IFileProperties
from mipac.types.emoji import EmojiPayload
from mipac.types.instance import (
    FeaturesPayload,
    InstancePayload,
    MetaPayload,
    OptionalInstance,
    OptionalMeta,
)
from mipac.types.note import GeoPayload, INote, INoteReaction, INoteRequired
from mipac.types.page import (
    AttachedFilePayload,
    EyeCatchingImagePayload,
    PageContentPayload,
    PageFilePayload,
    PagePayload,
    VariablePayload,
)
from mipac.types.poll import IPoll
from mipac.types.reaction import NoteReactionPayload

__all__ = (
    'ActiveUsersChartPayload',
    'DriveLocalChartPayload',
    'DriveRemoteChartPayload',
    'DriveChartPayload',
    'IFileProperties',
    'FolderPayload',
    'IDriveFile',
    'EmojiPayload',
    'INoteRequired',
    'INote',
    'GeoPayload',
    'IPoll',
    'FeaturesPayload',
    'INoteReaction',
    'MetaPayload',
    'InstancePayload',
    'OptionalInstance',
    'OptionalMeta',
    'PageContentPayload',
    'VariablePayload',
    'PageFilePayload',
    'EyeCatchingImagePayload',
    'AttachedFilePayload',
    'PagePayload',
    'NoteReactionPayload',
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, TypedDict

from mipac.types.ads import IAds
from mipac.types.emoji import ICustomEmoji

if TYPE_CHECKING:
    from mipac.types.emoji import EmojiPayload

__all__ = (
    'FeaturesPayload',
    'MetaPayload',
    'InstancePayload',
    'OptionalInstance',
    'OptionalMeta',
    'IInstanceLite',
    'IInstanceMeta',
)


class IInstanceLite(TypedDict):
    name: str
    software_name: str
    software_version: str
    icon_url: str
    favicon_url: str
    theme_color: str


class FeaturesPayload(TypedDict):
    registration: bool
    local_time_line: bool
    global_time_line: bool
    email_required_for_signup: bool
    elasticsearch: bool
    hcaptcha: bool
    recaptcha: bool
    object_storage: bool
    twitter: bool
    github: bool
    discord: bool
    service_worker: bool
    miauth: bool


class OptionalMeta(TypedDict, total=False):
    pinned_page: list[str]
    cache_remote_files: bool
    proxy_remote_files: bool
    require_setup: bool
    features: FeaturesPayload


class MetaPayload(OptionalMeta):
    maintainer_name: str
    maintainer_email: str
    version: str
    name: str
    uri: str
    description: str
    langs: list[str]
    tos_url: Optional[str]
    repository_url: str
    feedback_url: str
    secure: bool
    disable_registration: bool
    disable_local_timeline: bool
    disable_global_timeline: bool
    drive_capacity_per_local_user_mb: int
    drive_capacity_per_remote_user_mb: int
    email_required_for_signup: bool
    enable_hcaptcha: bool
    enable_recaptcha: bool
    recaptcha_site_key: str
    sw_publickey: str
    mascot_image_url: str
    error_image_url: str
    max_note_text_length: int
    emojis: list[EmojiPayload]
    ads: list
    enable_email: bool
    enable_twitter_integration: bool
    enable_github_integration: bool
    enable_discord_integration: bool
    enable_service_worker: bool
    translator_available: bool


class IInstanceMetaLiteRequired(TypedDict):
    version: str
    uri: str
    disable_registration: bool
    disable_local_timeline: bool
    disable_global_timeline: bool
    drive_capacity_per_local_user_mb: int
    drive_capacity_per_remote_user_mb: int
    enable_hcaptcha: bool
    max_note_text_length: int
    enable_email: bool
    enable_twitter_integration: bool
    enable_github_integration: bool
    enable_discord_integration: bool
    enable_service_worker: bool
    emojis: list[ICustomEmoji]
    mascot_image_url: str
    banner_url: str
    icon_url: str


class IInstanceMetaLite(IInstanceMetaLiteRequired, total=False):
    maintainer_name: str
    maintainer_email: str
    name: str
    description: str
    langs: list[str]
    tos_url: str
    tos_text_url: str
    announcements: dict[str, Any]
    hcaptcha_site_key: str
    enable_recaptcha: bool
    recaptcha_siteKey: str
    sw_publickey: str
    ads: list[IAds]  # v12 only


class IInstanceMeta(IInstanceMetaLite):
    features: FeaturesPayload


class OptionalInstance(TypedDict, total=False):
    host: str
    software_name: str
    software_version: str
    icon_url: str
    favicon_url: str
    theme_color: str


class InstancePayload(OptionalInstance, MetaPayload):
    meta: MetaPayload



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.types.emoji import ICustomEmoji

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ('CustomEmoji',)


class CustomEmoji:
    """
    Attributes
    ----------
    id : Optional[str]
        絵文字のID
    aliases : Optional[list[str]]
        絵文字のエイリアス
    name : Optional[str]
        絵文字の名前
    category : Optional[str]
        絵文字のカテゴリ
    host : Optional[str]
        絵文字のホスト
    url : Optional[str]
        絵文字のURL
    """

    __slots__ = (
        '__emoji',
        '__client',
    )

    def __init__(self, emoji: ICustomEmoji, *, client: ClientActions):
        self.__emoji: ICustomEmoji = emoji
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__emoji['id']

    @property
    def aliases(self) -> list[str]:
        return self.__emoji['aliases']

    @property
    def name(self) -> str:
        return self.__emoji['name']

    @property
    def category(self) -> str:
        return self.__emoji['category']

    @property
    def url(self) -> str:
        return self.__emoji['url']



from typing import Literal

from mipac.models.lite.instance import LiteInstance
from mipac.types.emoji import ICustomEmojiLite
from mipac.types.user import IUserLite


class LiteUser:
    __slots__ = ('__user',)

    def __init__(self, user: IUserLite) -> None:
        self.__user: IUserLite = user

    @property
    def id(self) -> str:
        return self.__user['id']

    @property
    def username(self) -> str:
        return self.__user['username']

    @property
    def host(self) -> str | None:
        return self.__user['host']

    @property
    def name(self) -> str:
        return self.__user['name']

    @property
    def online_status(
        self,
    ) -> Literal['online', 'active', 'offline', 'unknown']:
        return self.__user['online_status']

    @property
    def avatar_url(self) -> str:
        return self.__user['avatar_url']

    @property
    def avatar_blurhash(self) -> str:
        return self.__user['avatar_blurhash']

    @property
    def emojis(self) -> list[ICustomEmojiLite]:  # TODO: ちゃんとモデルにする
        return self.__user['emojis']

    @property
    def instance(self) -> LiteInstance:
        return LiteInstance(self.__user['instance'])



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict


if TYPE_CHECKING:
    from mipac.types.user import IUserLite

__all__ = ('NoteReactionPayload',)


class NoteReactionPayload(TypedDict):
    id: str
    created_at: str
    user: IUserLite
    type: str



from mipac.core.models.chart import (
    RawActiveUsersChart,
    RawDriveChart,
    RawDriveLocalChart,
    RawDriveRemoteChart,
)

__all__ = (
    'RawActiveUsersChart',
    'RawDriveRemoteChart',
    'RawDriveLocalChart',
    'RawDriveChart',
)



from __future__ import annotations
from typing import TYPE_CHECKING
from mipac.models.lite.user import LiteUser
from mipac.types.follow import IFollowRequest

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions


class FollowRequest:
    def __init__(
        self, follow_request: IFollowRequest, *, client: ClientActions
    ) -> None:
        self.__follow_request = follow_request
        self.__client = client

    @property
    def id(self) -> str:
        return self.__follow_request['id']

    @property
    def follower(self) -> LiteUser:
        return LiteUser(self.__follow_request['follower'])

    @property
    def followee(self) -> LiteUser:
        return LiteUser(self.__follow_request['followee'])



from typing import Any, Literal, Optional, TypedDict

from .drive import IDriveFile
from .emoji import ICustomEmojiLite
from .poll import IPoll
from .user import IUserLite


class GeoPayload(TypedDict):
    """
    衛星情報
    """

    coordinates: Optional[list[Any]]
    altitude: Optional[int]
    accuracy: Optional[int]
    altitude_accuracy: Optional[int]
    heading: Optional[int]
    speed: Optional[int]


class INoteRequired(TypedDict):
    id: str
    created_at: str
    text: str | None
    cw: str | None
    user: IUserLite
    user_id: str
    reply_id: str
    renote_id: str
    files: list[IDriveFile]
    file_ids: list[str]
    visibility: Literal['public', 'home', 'followers', 'specified']
    reactions: dict[str, int]
    renote_count: int
    replies_count: int
    emojis: list[ICustomEmojiLite]


class INote(INoteRequired, total=False):
    """
    note object
    """

    renote: 'INote'
    reply: 'INote'
    visible_user_ids: list[str]
    local_only: bool
    my_reaction: str
    uri: str
    url: str
    is_hidden: bool
    poll: IPoll


class INoteReaction(TypedDict):
    id: str
    created_at: str
    user: IUserLite
    type: str



from __future__ import annotations

from typing import TYPE_CHECKING
from mipac.abc.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.follow import FollowRequest
from mipac.types.follow import IFollowRequest

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions


class MyActions(AbstractAction):
    def __init__(self, session: HTTPClient, client: ClientActions):
        self.__session = session
        self.__client = client

    async def fetch_follow_requests(self):
        res: list[IFollowRequest] = await self.__session.request(
            Route('POST', '/api/following/requests/list'),
            auth=True,
            lower=True,
        )
        return [FollowRequest(i, client=self.__client) for i in res]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from mipac.http import HTTPClient

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions

__all__ = ('AbstractManager',)


class AbstractManager(ABC):
    @abstractmethod
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.models.lite.instance import LiteInstanceMeta
from mipac.types.instance import IInstanceMeta

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ('InstanceMeta',)


class InstanceMeta(LiteInstanceMeta):
    def __init__(
        self, instance: IInstanceMeta, *, client: ClientActions
    ) -> None:
        super().__init__(instance, client=client)
        self.__features = instance['features']

    @property
    def registration(self) -> bool:
        return self.__features['registration']

    @property
    def local_time_line(self) -> bool:
        return self.__features['local_time_line']

    @property
    def global_time_line(self) -> bool:
        return self.__features['global_time_line']

    @property
    def email_required_for_signup(self) -> bool:
        return self.__features['email_required_for_signup']

    @property
    def elasticsearch(self) -> bool:
        return self.__features['elasticsearch']

    @property
    def hcaptcha(self) -> bool:
        return self.__features['hcaptcha']

    @property
    def recaptcha(self) -> bool:
        return self.__features['recaptcha']

    @property
    def object_storage(self) -> bool:
        return self.__features['object_storage']

    @property
    def twitter(self) -> bool:
        return self.__features['twitter']

    @property
    def github(self) -> bool:
        return self.__features['github']

    @property
    def discord(self) -> bool:
        return self.__features['discord']

    @property
    def service_worker(self) -> bool:
        return self.__features['service_worker']

    @property
    def miauth(self) -> bool:
        return self.__features['miauth']



# noqa: F403, F401

from .ad import *
from .chart import *
from .chat import *
from .client import *
from .drive import *
from .emoji import *
from .favorite import *
from .file import *
from .follow import *
from .moderator import *
from .note import *
from .page import *
from .reaction import *
from .user import *



from typing import Optional


class MiFile:
    __slots__ = (
        'path',
        'file_id',
        'name',
        'folder_id',
        'comment',
        'is_sensitive',
        'force',
    )

    def __init__(
        self,
        path: Optional[str] = None,
        file_id: Optional[str] = None,
        name: Optional[str] = None,
        folder_id: Optional[str] = None,
        comment: Optional[str] = None,
        is_sensitive: bool = False,
        force: bool = False,
    ):
        """
        Parameters
        ----------
        path : Optional[str], default=None
            path to a local file
        file_id : Optional[str], default=None
            ID of the file that exists on the drive
        name Optional[str], default=None
            file name
        folder_id : Optional[str], default=None
            Folder ID
        comment : Optional[str], default=None
            Comments on files
        is_sensitive : Optional[str], default=None
            Whether this item is sensitive
        force : bool, default=False
            Whether to force overwriting even if it already exists on the drive
        """
        self.path = path
        self.file_id = file_id
        self.name = name
        self.folder_id = folder_id
        self.comment = comment
        self.is_sensitive = is_sensitive
        self.force = force



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Literal, Optional


from mipac.exception import NotExistRequiredData
from mipac.models.lite.user import LiteUser
from mipac.models.poll import Poll

if TYPE_CHECKING:
    from mipac.actions.note import NoteActions
    from mipac.manager.client import ClientActions
    from mipac.models.user import UserDetailed
    from mipac.types.drive import IDriveFile
    from mipac.types.emoji import ICustomEmojiLite
    from mipac.types.note import INote, INoteReaction

__all__ = (
    'Note',
    'Follow',
    'Header',
    'NoteReaction',
)


class Follow:
    def __init__(self, data):
        self.id: Optional[str] = data.get('id')
        self.created_at: Optional[datetime] = datetime.strptime(
            data['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        ) if data.get('created_at') else None
        self.type: Optional[str] = data.get('type')
        self.user: Optional[UserDetailed] = data.get('user')

    async def follow(self) -> tuple[bool, Optional[str]]:
        """
        ユーザーをフォローします
        Returns
        -------
        bool
            成功ならTrue, 失敗ならFalse
        str
            実行に失敗した際のエラーコード
        """

        if self.id:
            raise NotExistRequiredData('user_idがありません')
        return await self._state.user.follow.add(user_id=self.id)

    async def unfollow(self, user_id: Optional[str] = None) -> bool:
        """
        与えられたIDのユーザーのフォローを解除します

        Parameters
        ----------
        user_id : Optional[str] = None
            フォローを解除したいユーザーのID

        Returns
        -------
        status
            成功ならTrue, 失敗ならFalse
        """

        if user_id is None:
            user_id = self.user.id
        return await self._state.user.follow.remove(user_id)


class Header:
    def __init__(self, data):
        self.id = data.get('id')
        self.type = data.get('type')


class NoteReaction:
    __slots__ = ('__reaction',)

    def __init__(self, reaction: INoteReaction):
        self.__reaction: INoteReaction = reaction

    @property
    def id(self) -> str | None:
        return self.__reaction['id']

    @property
    def created_at(self) -> datetime | None:
        return (
            datetime.strptime(
                self.__reaction['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
            )
            if 'created_at' in self.__reaction
            else None
        )

    @property
    def type(self) -> str | None:
        return self.__reaction['type']

    @property
    def user(self) -> LiteUser:
        return LiteUser(self.__reaction['user'])


class Note:
    """
    Noteモデル

    Parameters
    ----------
    note: INote
        アクションを持たないNoteクラス
    client: ClientActions
    """

    def __init__(self, note: INote, client: ClientActions):
        self.__note = note
        self._client: ClientActions = client

    @property
    def id(self) -> str:
        """
        ユーザーのID

        Returns
        -------
        str
            ユーザーのID
        """
        return self.__note['id']

    @property
    def created_at(self) -> datetime:
        return datetime.strptime(
            self.__note['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )

    @property
    def content(self) -> str | None:
        return self.__note['text']

    @property
    def cw(self) -> str | None:
        return self.__note['cw']

    @property
    def user_id(self) -> str:
        return self.__note['user_id']

    @property
    def author(self) -> LiteUser:
        return LiteUser(self.__note['user'])

    @property
    def reply_id(self) -> str:
        return self.__note['reply_id']

    @property
    def renote_id(self) -> str:
        return self.__note['renote_id']

    @property
    def files(self) -> list[IDriveFile]:  # TODO: モデルに
        return self.__note['files']

    @property
    def file_ids(self) -> list[str]:
        return self.__note['file_ids']

    @property
    def visibility(
        self,
    ) -> Literal['public', 'home', 'followers', 'specified']:
        return self.__note['visibility']

    @property
    def reactions(self) -> dict[str, int]:
        return self.__note['reactions']

    @property
    def renote_count(self) -> int:
        return self.__note['renote_count']

    @property
    def replies_count(self) -> int:
        return self.__note['replies_count']

    @property
    def emojis(self) -> list[ICustomEmojiLite]:  # TODO: モデルに
        return self.__note['emojis']

    @property
    def renote(self) -> 'Note' | None:
        return (
            Note(note=self.__note['renote'], client=self._client)
            if 'renote' in self.__note
            else None
        )

    @property
    def reply(self) -> 'Note' | None:
        return (
            Note(note=self.__note['renote'], client=self._client)
            if 'renote' in self.__note
            else None
        )

    @property
    def visible_user_ids(self) -> list[str]:
        return (
            self.__note['visible_user_ids']
            if 'visible_user_ids' in self.__note
            else []
        )

    @property
    def local_only(self) -> bool:
        return (
            self.__note['local_only'] if 'local_only' in self.__note else False
        )

    @property
    def my_reaction(self) -> str | None:
        return (
            self.__note['my_reaction']
            if 'my_reaction' in self.__note
            else None
        )

    @property
    def uri(self) -> str | None:
        return self.__note['uri'] if 'uri' in self.__note else None

    @property
    def url(self) -> str | None:
        return self.__note['url'] if 'url' in self.__note else None

    @property
    def is_hidden(self) -> bool:
        return (
            self.__note['is_hidden'] if 'is_hidden' in self.__note else False
        )

    @property
    def poll(self) -> Poll | None:
        return (
            Poll(self.__note['poll'], client=self._client)
            if 'poll' in self.__note
            else None
        )

    @property
    def action(self) -> NoteActions:
        """
        ノートに対するアクション

        Returns
        -------
        NoteActions
        """
        return self._client._create_note_instance(self.id).action



from __future__ import annotations

import sys
from typing import Any, Literal, TypeVar

import aiohttp

from mipac import __version__
from mipac.exception import APIError
from mipac.types.endpoints import ENDPOINTS
from mipac.types.user import IUserDetailed
from mipac.util import remove_dict_empty, upper_to_lower, _from_json


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()
R = TypeVar('R')


async def json_or_text(response: aiohttp.ClientResponse):
    text = await response.text(encoding='utf-8')
    try:
        if 'application/json' in response.headers['Content-Type']:
            return _from_json(text)
    except KeyError:
        pass


class Route:
    def __init__(self, method: Literal['GET', 'POST'], path: ENDPOINTS):
        self.path: str = path
        self.method: str = method


class HTTPClient:
    def __init__(self, url: str, token: str) -> None:
        user_agent = (
            'Misskey Bot (https://github.com/yupix/MiPA {0})'
            + 'Python/{1[0]}.{1[1]} aiohttp/{2}'
        )
        self.user_agent = user_agent.format(
            __version__, sys.version_info, aiohttp.__version__
        )
        self.__session: aiohttp.ClientSession = MISSING
        self.__url: str = url
        self.__token: str = token

    @property
    def session(self) -> aiohttp.ClientSession:
        return self.__session

    async def request(self, route: Route, **kwargs) -> R:
        headers: dict[str, str] = {
            'User-Agent': self.user_agent,
        }

        is_lower = kwargs.pop('lower') if kwargs.get('lower') else False

        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['json'] = kwargs.pop('json')

        if kwargs.get('auth') and kwargs.pop('auth'):
            key = (
                'json' if 'json' in kwargs or 'data' not in kwargs else 'data'
            )
            if not kwargs.get(key):
                kwargs[key] = {}
            kwargs[key]['i'] = self.__token

        replace_list = kwargs.pop('replace_list', {})

        for i in ('json', 'data'):
            if kwargs.get(i):
                kwargs[i] = remove_dict_empty(kwargs[i])

        async with self.__session.request(
            route.method, self.__url + route.path, **kwargs
        ) as res:
            data = await json_or_text(res)
            if is_lower:
                if isinstance(data, list):
                    data = [
                        upper_to_lower(i, replace_list=replace_list)
                        for i in data
                    ]
            if 300 > res.status >= 200:
                return data  # type: ignore
            if 511 > res.status >= 300:
                raise APIError(data, res.status)
            raise APIError('HTTP ERROR')

    async def close_session(self) -> None:
        await self.__session.close()

    async def login(self) -> IUserDetailed:
        self.__session = aiohttp.ClientSession()
        data: IUserDetailed = await self.request(
            Route('POST', '/api/i'), auth=True
        )
        return data



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mipac.abc.manager import AbstractManager
from mipac.actions.note import NoteActions
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions


class NoteManager(AbstractManager):
    """User behavior for notes"""

    def __init__(
        self,
        note_id: Optional[str] = None,
        *,
        session: HTTPClient,
        client: ClientActions
    ):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client
        self.__note_id = note_id

    @property
    def action(self) -> NoteActions:
        return NoteActions(
            note_id=self.__note_id,
            session=self.__session,
            client=self.__client,
        )

    async def get(
        self,
        local: bool = True,
        reply: bool = False,
        renote: bool = True,
        with_files: bool = False,
        poll: bool = True,
        limit: int = 10,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
    ):
        data = {
            'local': local,
            'reply': reply,
            'renote': renote,
            'withFiles': with_files,
            'poll': poll,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
        }
        await self.__session.request(
            Route('POST', '/api/notes'), json=data, auth=True, lower=True
        )



from typing import TypedDict

__all__ = ('IPoll', 'IPollChoice')


class IPollChoice(TypedDict):
    is_voted: bool
    text: str
    votes: int


class IPoll(TypedDict, total=False):
    """
    アンケート情報
    """

    multiple: bool
    expires_at: int
    choices: list[IPollChoice]



class NotExistRequiredData(Exception):
    """クラスの中に必要なデータが不足している"""


class ParameterError(Exception):
    """引数に関するエラー"""


class NotSupportedError(Exception):
    """特定のForkでサポートしていない場合のエラー"""


class APIError(Exception):
    """APIのエラー"""



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from mipac.models.lite.user import LiteUser
from mipac.models.note import Note
from mipac.types.page import IPage
from mipac.types.user import IFollowRequest, IUserDetailed, IUserDetailedField

if TYPE_CHECKING:
    from mipac.actions.user import UserActions
    from mipac.manager.client import ClientActions

__all__ = ('UserDetailed', 'FollowRequest')


class FollowRequest:
    def __init__(self, request: IFollowRequest, *, client: ClientActions):
        self.__request: IFollowRequest = request
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__request['id']

    @property
    def follower(self) -> LiteUser:
        return LiteUser(self.__request['follower'])

    @property
    def followee(self) -> LiteUser:
        return LiteUser(self.__request['followee'])


class UserDetailed(LiteUser):
    __slots__ = (
        '__detail',
        '__client',
    )

    def __init__(self, user: IUserDetailed, *, client: ClientActions):
        super().__init__(user=user)
        self.__detail = user
        self.__client: ClientActions = client

    @property
    def fields(self) -> list[IUserDetailedField]:
        return self.__detail['fields']

    @property
    def followers_count(self) -> int:
        return self.__detail['followers_count']

    @property
    def following_count(self) -> int:
        return self.__detail['following_count']

    @property
    def has_pending_follow_request_from_you(self) -> bool:
        return self.__detail['has_pending_follow_request_from_you']

    @property
    def has_pending_follow_request_to_you(self) -> bool:
        return self.__detail['has_pending_follow_request_to_you']

    @property
    def is_admin(self) -> bool:
        return self.__detail['is_admin']

    @property
    def is_blocked(self) -> bool:
        return self.__detail['is_blocked']

    @property
    def is_blocking(self) -> bool:
        return self.__detail['is_blocking']

    @property
    def is_bot(self) -> bool:
        return self.__detail['is_bot']

    @property
    def is_cat(self) -> bool:
        return self.__detail['is_cat']

    @property
    def is_followed(self) -> bool:
        return self.__detail['is_followed']

    @property
    def is_following(self) -> bool:
        return self.__detail['is_following']

    @property
    def is_locked(self) -> bool:
        return self.__detail['is_locked']

    @property
    def is_moderator(self) -> bool:
        return self.__detail['is_moderator']

    @property
    def is_muted(self) -> bool:
        return self.__detail['is_muted']

    @property
    def is_silenced(self) -> bool:
        return self.__detail['is_silenced']

    @property
    def is_suspended(self) -> bool:
        return self.__detail['is_suspended']

    @property
    def public_reactions(self) -> bool:
        return self.__detail['public_reactions']

    @property
    def security_keys(self) -> bool:
        return self.__detail['security_keys']

    @property
    def two_factor_enabled(self) -> bool:
        return self.__detail['two_factor_enabled']

    @property
    def notes_count(self) -> int:
        return self.__detail['notes_count']

    @property
    def pinned_note_ids(self) -> list[str]:
        return self.__detail['pinned_note_ids']

    @property
    def pinned_notes(self) -> list[Note]:
        return [
            Note(i, client=self.__client)
            for i in self.__detail['pinned_notes']
        ]

    @property
    def banner_blurhash(self) -> str | None:
        return self.__detail.get('banner_blurhash')

    @property
    def banner_color(self) -> str | None:
        return self.__detail.get('banner_color')

    @property
    def banner_url(self) -> str | None:
        return self.__detail.get('banner_url')

    @property
    def birthday(self) -> str | None:
        return self.__detail.get('birthday')

    @property
    def created_at(self) -> str | None:
        return self.__detail.get('created_at')

    @property
    def description(self) -> str | None:
        return self.__detail.get('description')

    @property
    def ff_visibility(
        self,
    ) -> Literal['public', 'followers', 'private'] | None:
        return self.__detail.get('ff_visibility')

    @property
    def lang(self) -> str | None:
        return self.__detail.get('lang')

    @property
    def last_fetched_at(self) -> str | None:
        return self.__detail.get('last_fetched_at')

    @property
    def location(self) -> str | None:
        return self.__detail.get('location')

    @property
    def pinned_page(self) -> IPage | None:
        return self.__detail.get('pinned_page')

    @property
    def pinned_page_id(self) -> str | None:
        return self.__detail.get('pinned_page_id')

    @property
    def updated_at(self) -> str | None:
        return self.__detail.get('updated_at')

    @property
    def uri(self) -> str | None:
        return self.__detail.get('uri')

    @property
    def url(self) -> str | None:
        return self.__detail.get('url')

    @property
    def action(self) -> UserActions:
        return self.__client._create_user_instance(self).action



from typing import Optional

from mipac.config import Config
from mipac.http import HTTPClient
from mipac.manager.client import ClientActions


class Client:
    def __init__(
        self, url: str, token: str, *, config: Optional[Config] = None
    ):
        self.__url: str = url
        self.__token: str = token
        self.__config: Config = config or Config()
        self.http: HTTPClient = HTTPClient(url, token)

    @property
    def api(self):
        return ClientActions(self.http, self.__config)

    async def close_session(self):
        await self.http.close_session()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional


from mipac.exception import NotExistRequiredData, ParameterError
from mipac.http import HTTPClient, Route
from mipac.manager.note import NoteManager
from mipac.util import (
    cache,
    check_multi_arg,
    remove_dict_empty,
)

if TYPE_CHECKING:
    from mipac.manager.client import ClientActions
    from mipac.models.note import Note
    from mipac.models.user import UserDetailed

__all__ = ['UserActions']


class UserActions:
    def __init__(
        self,
        session: HTTPClient,
        client: ClientActions,
        user: Optional[UserDetailed] = None,
    ):
        self.__session: HTTPClient = session
        self.__user: Optional[UserDetailed] = user
        self.__client: ClientActions = client
        self.note: NoteManager = NoteManager(session=session, client=client)

    async def get_me(self) -> UserDetailed:
        """
        ログインしているユーザーの情報を取得します
        """

        res = await self.__session.request(Route('POST', '/api/i'), auth=True)
        return UserDetailed(res, client=self.__client)  # TODO: 自分用のクラスに変更する

    @cache(group='get_user')
    async def get(
        self,
        user_id: Optional[str] = None,
        username: Optional[str] = None,
        host: Optional[str] = None,
    ) -> UserDetailed:
        """
        ユーザーのプロフィールを取得します。一度のみサーバーにアクセスしキャッシュをその後は使います。
        fetch_userを使った場合はキャッシュが廃棄され再度サーバーにアクセスします。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        UserDetailed
            ユーザー情報
        """

        field = remove_dict_empty(
            {'userId': user_id, 'username': username, 'host': host}
        )
        data = await self.__session.request(
            Route('POST', '/api/users/show'), json=field, auth=True, lower=True
        )
        return UserDetailed(data, client=self.__client)

    @cache
    async def fetch(
        self,
        user_id: Optional[str] = None,
        username: Optional[str] = None,
        host: Optional[str] = None,
    ) -> UserDetailed:
        """
        サーバーにアクセスし、ユーザーのプロフィールを取得します。基本的には get_userをお使いください。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        UserDetailed
            ユーザー情報
        """
        if not check_multi_arg(user_id, username):
            raise ParameterError('user_id, usernameどちらかは必須です')

        field = remove_dict_empty(
            {'userId': user_id, 'username': username, 'host': host}
        )
        data = await self.__session.request(
            Route('POST', '/api/users/show'), json=field, auth=True, lower=True
        )
        return UserDetailed(data, client=self.__client)

    async def get_notes(
        self,
        user_id: Optional[str] = None,
        include_replies: bool = True,
        limit: int = 10,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        since_date: int = 0,
        until_date: int = 0,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: Optional[list[str]] = None,
        exclude_nsfw: bool = True,
    ) -> list[Note]:

        if check_multi_arg(user_id, self.__user):
            raise ParameterError('user_idがありません')

        user_id = user_id or self.__user and self.__user.id
        data = {
            'userId': user_id,
            'includeReplies': include_replies,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'sinceDate': since_date,
            'untilDate': until_date,
            'includeMyRenotes': include_my_renotes,
            'withFiles': with_files,
            'fileType': file_type,
            'excludeNsfw': exclude_nsfw,
        }
        res = await self.__session.request(
            Route('POST', '/api/users/notes'), json=data, auth=True, lower=True
        )
        return [Note(i, client=self.__client) for i in res]

    def get_mention(self, user: Optional[UserDetailed] = None) -> str:
        """
        Get mention name of user.

        Parameters
        ----------
        user : Optional[User], default=None
            メンションを取得したいユーザーのオブジェクト

        Returns
        -------
        str
            メンション
        """

        user = user or self.__user

        if user is None:
            raise NotExistRequiredData('Required parameters: user')
        return (
            f'@{user.name}@{user.host}' if user.instance else f'@{user.name}'
        )



from typing import TypedDict


class IChannel(TypedDict):
    id: str
    created_at: str
    last_noted_at: str
    name: str
    description: str | None
    banner_url: str | None
    notes_count: int
    users_count: int
    is_following: bool
    user_id: str
    has_unread_note: bool



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.types.channel import IChannel

if TYPE_CHECKING:
    from mipac.manager import ClientActions


class Channel:
    def __init__(self, channel: IChannel, *, client: ClientActions) -> None:
        self.__channel: IChannel = channel
        self.__client: ClientActions = client

    @property
    def id(self) -> str:
        return self.__channel['id']

    @property
    def created_at(self) -> str:
        return self.__channel['created_at']

    @property
    def last_noted_at(self) -> str:
        return self.__channel['last_noted_at']

    @property
    def name(self) -> str:
        return self.__channel['name']

    @property
    def description(self) -> str | None:
        return self.__channel['description']

    @property
    def banner_url(self) -> str | None:
        return self.__channel['banner_url']

    @property
    def notes_count(self) -> int:
        return self.__channel['notes_count']

    @property
    def users_count(self) -> int:
        return self.__channel['users_count']

    @property
    def is_following(self) -> bool:
        return bool(self.__channel['is_following'])

    @property
    def user_id(self) -> str:
        return self.__channel['user_id']

    @property
    def has_unread_note(self) -> bool:
        return self.__channel['has_unread_note']



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abc.manager import AbstractManager
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.client import ClientActions

__all__ = ('AdminAdvertisingManager',)


class AdminAdvertisingManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientActions):
        self.__session: HTTPClient = session
        self.__client: ClientActions = client

    async def create(
        self,
        url: str,
        memo: str,
        place: str,
        priority: str,
        ratio: str,
        expires_at: int,
        image_url: str,
    ):
        data = {
            'url': url,
            'memo': memo,
            'place': place,
            'priority': priority,
            'ratio': ratio,
            'expires_at': expires_at,
            'image_url': image_url,
        }
        return await self.__session.request(
            Route('POST', '/api/admin/ad/create'),
            json=data,
            auth=True,
            lower=True,
        )
